from .parser import Parser
from .handle_invalid_scenarios import ParserError


def ganit():
    p = Parser()
    print("Type quit to exit")
    while True:
        print()
        exp = input("Ganit>>> ")
        if exp == 'quit':
            break
        try:
            print("Prescanned Expression:", p.prescan(exp))
            print("Postfix Notation:", p.convert(exp))
            print("Evaluation Result:", p.evaluate(exp))
        except ParserError as e:
            print("ParserError:",str(e))



# A python utility to evaluate an Infix expression.
from .utils import *
from .variables_operations import *
from .all_operators import all_operators
from .handle_invalid_scenarios import *
from .converter import convert
def evaluate(postfix, variables):
    if is_float(postfix) or is_int(postfix):
        return postfix
    if variables is not None:
        variables = {**default_variables, **variables}
    else:
        variables = default_variables
    stack=[]
    ternary = []
    elements = postfix.split()

    def check_postfix():
        counter = 0
        for element in elements:
            if is_valid_operand(element, variables):
                counter += 1
                if counter < 0:
                    break
            elif element == ',' or element == "(" or element == ")":
                handle_invalid_scenario("Invalid Postfix Notation: Contains ',' or '(' or ')'")
            elif element in all_operators:
                operator_definition = all_operators[element]
                numOfOperand = operator_definition["argument"]
                counter -= numOfOperand
                if counter < 0:
                    break    
                counter += 1
            else:
                handle_invalid_scenario("Unknown Symbol in expression: " + element)
        if counter < 0:
            handle_invalid_scenario("Invalid Postfix Notation: Causes Stack Underflow - " + postfix)
        elif counter != 1:
            handle_invalid_scenario("Invalid Postfix Notation: Stack not 1")
                
        return True


    def handle_expression(argument):
        expr = argument.get("exp")
        expVariables = argument.get("variables", None)
        expConvert = argument.get("convert", None)
        if isinstance(expr, dict):
            return handle_expression(expr)
        else:
            if expConvert is None or expConvert==True:
                convertedExp = convert(expr, expVariables)
                value = evaluate(convertedExp, expVariables)
            else:
                value = evaluate(expr, expVariables)
                
        return value
        


    def handle_operation(operator):
        nonlocal stack,ternary
        arguments = []
        
        operator_definition = all_operators[operator]
        numOfOperand = operator_definition["argument"]
        for x in range(numOfOperand):
            argument  = stack.pop()
            if argument in variables:
                argument = variables[argument]
            if isinstance(argument, dict):
                value =handle_expression(argument)
            elif isinstance(argument, bool):
                value=bool(argument)
            elif is_float(argument):
                value = float(argument)
            elif is_int(argument):
                value = int(argument)
            elif isinstance(argument, str):
                convertedExp = convert(argument, None)
                value = evaluate(convertedExp, None)
            else:
                value = argument
            check = is_int(argument) or is_float(argument)
            if check == False and value is None:
                handle_invalid_scenario("Value Error: invalid value passed: " + argument)
            if isinstance(argument, bool):
                arguments.append(bool(value))
            else:
                arguments.append(float(value))
        
        # Handle ternary operation
        if operator == '?':
            ternary.append(True)

        if operator == ':':
            if len(ternary) <= 0:
                handle_invalid_scenario("Invalid operation ':', no matching '?' found")
            else:
                currTernary = ternary.pop()
                if currTernary is None:
                    handle_invalid_scenario("Invalid operation with :")
                
        result = perform_operation(operator, arguments)
        stack.append(result)

    check_postfix()
    for element in elements:
        if is_valid_operand(element, variables):
            if element in variables:
                value  = variables[element]
                if isinstance(value, dict):
                    value = handle_expression(value)
            elif isinstance(element, bool):
                value = bool(element)
            elif is_int(element):
                value = int(element)
            elif is_float(element):
                value = float(element)
            
            stack.append(value)
        elif element == ',':
            handle_invalid_scenario("Invalid expression: contains comma")
        elif element in all_operators:
            handle_operation(element)
        else:
            handle_invalid_scenario("Unknown Symbol in expression: " + element)
    
    result = stack[0]
    
    if isinstance(result, str) and result in variables:
        result = variables[result]
    return result



#Prescan the expression to correctly identify the negation and subtraction
from .handle_invalid_scenarios import *
from .utils import is_float
from .utils import is_int
from .all_operators import all_operators


def pre_scan(exp):
    if is_float(exp) or is_int(exp):
        return exp
    if isinstance(exp, dict):
        exp = exp["exp"]
    exp = format(exp)
    exp = exp.replace(" ", "")
    #Changes required to perform shunting yard algorithm on expression
    # Change != to space whhich will be converted back to ! after all the conversion happens
    exp = exp.replace("!=", " ")
    # Change == to =
    exp = exp.replace("==", "=")
    # Change <= to #
    exp = exp.replace("<=", "#")
    # Change >= to @
    exp = exp.replace(">=", "@")
    expr = ''
    segment = False

    for idx, letter in enumerate(exp):
        if letter == '-':
            #Write all the logic for negation and subtraction
            if idx == 0:
                expr += "0" + letter
            elif exp[idx - 1] == "(":
                expr += "0" + letter
            elif exp[idx - 1] in all_operators and exp[idx - 1] == "-":
                if expr[-1] == '-':
                    expr = expr[:-1] + '+'
                else:
                    expr = expr[:-1] + '-'
            elif exp[idx - 1] in all_operators and exp[idx - 1] != ")":
                expr += "(0" + letter
                segment = True
            else:
                expr += letter
        elif letter == '+':
            #Write all the logic for positive number and addition
            if idx == 0:
                continue
            elif exp[idx - 1] == "(":
                continue
            elif exp[idx - 1] in all_operators and exp[idx - 1] == "+":
                continue
            elif exp[idx - 1] in all_operators and exp[idx - 1] != ")":
                continue
            else:
                expr += letter
        elif letter == '!':
            if exp[idx + 1] == '=':
                expr += '!'
                idx += 1
            else:
                expr += '~'
        else:
            check = True
            if letter in all_operators and segment:
                expr += ")"
                segment = False
            elif segment and idx == len(exp) - 1:
                expr += letter + ")"
                check = False
            if check:
                expr += letter
    
    expr = expr.replace(" ", "!")
    exp = expr

    return exp


#End of prescan



from math import *
from .utils import *
from .handle_invalid_scenarios import handle_invalid_scenario


def roundOp(arg):
    if is_int(arg[0]):
        return round(arg[1], int(arg[0]))
    else:
        handle_invalid_scenario("Rounding takes only 'int' as second argument")


def fact(arg):
    if is_int(arg[0]) == False or int(arg[0]) < 0:
        handle_invalid_scenario(
            "Value Error: " + format(arg[0]) +
            ". Factorial calculation can only be done for integers greater than 0."
        )
    return (factorial(arg[0]))


def gcdOp(arg):
    if (is_int(arg[0]) == False) or (is_int(arg[1]) == False):
        handle_invalid_scenario("Value Error: " + format(arg[0]) + " , " +
                                format(arg[1]) +
                                ". GCD can only be determined for integers.")
    return gcd(int(arg[1]), int(arg[0]))


def square(arg):
    if arg[0] < 0:
        handle_invalid_scenario("Square root only works for positive numbers")
    return sqrt(arg[0])


algebraic = {
    "abs": lambda arg: abs(arg[0]),
    "ceil": lambda arg: ceil(arg[0]),
    "floor": lambda arg: floor(arg[0]),
    "round": lambda arg: roundOp(arg),
    "factorial": lambda arg: fact(arg),
    "gcd": lambda arg: gcdOp(arg),
    "exp": lambda arg: exp(arg[0]),
    "pow": lambda arg: pow(arg[1], arg[0]),
    "sqrt": lambda arg: square(arg),
    "log": lambda arg: log10(arg[0]),
    "ln": lambda arg: log(arg[0]),
    "log2": lambda arg: log2(arg[0]),
    "Log": lambda arg: log(arg[1], arg[0])
}



class ParserError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return (repr(self.value))


def handle_invalid_scenario(msg):
    raise ParserError(msg)



from math import *
from .utils import *
from .handle_invalid_scenarios import handle_invalid_scenario


def question(arg):
    if isinstance(arg[1], bool) and arg[1] == True:
        return arg[0]
    elif isinstance(arg[1], bool) and arg[1] == False:
        return arg[1]
    else:
        handle_invalid_scenario("'?:' only works for boolean condition")


def colon(arg):
    if isinstance(arg[1], bool) and arg[1] == False:
        return arg[0]
    else:
        return arg[1]


def inverse(arg):
    if not is_bool(arg[0]):
        handle_invalid_scenario("'!' only works for boolean")
    return not arg[0]


def orOp(arg):
    if isinstance(arg[1], bool) and isinstance(arg[0], bool):
        return arg[0] or arg[1]
    else:
        handle_invalid_scenario("'|' only works for boolean conditions")


def andOp(arg):
    if isinstance(arg[1], bool) and isinstance(arg[0], bool):
        return arg[0] and arg[1]
    else:
        handle_invalid_scenario("'&' only works for boolean conditions")


logical = {
    "?": lambda arg: question(arg),
    ":": lambda arg: colon(arg),
    ">": lambda arg: arg[1] > arg[0],
    "<": lambda arg: arg[1] < arg[0],
    "=": lambda arg: arg[1] == arg[0],
    "@": lambda arg: arg[1] >= arg[0],
    "#": lambda arg: arg[1] <= arg[0],
    "!": lambda arg: arg[1] != arg[0],
    "~": lambda arg: inverse(arg),
    '|': lambda arg: orOp(arg),
    '&': lambda arg: andOp(arg)
}


basic_arithmetic = {
    "+": lambda arg: arg[1] + arg[0],
    "-": lambda arg: arg[1] - arg[0],
    "*": lambda arg: arg[1] * arg[0],
    "/": lambda arg: arg[1] / arg[0],
    "%": lambda arg: arg[1] % arg[0],
    "^": lambda arg: arg[1]**arg[0]
}


# A python utility to evaluate an Infix expression.
from .utils import *
from .variables_operations import default_variables
from .all_operators import all_operators
from .handle_invalid_scenarios import *
def convert(exp, variables):
    if is_float(exp) or is_int(exp):
        return exp
    if variables is not None:
        variables = {**default_variables, **variables}
    else:
        variables = default_variables
    operands = []
    stack = []
    operand = ""
    postFix = ""
    stack.append('(')

    def add_operand():
        nonlocal postFix, operand
        operand = operand.strip()
        if is_valid_operand(operand, variables):
            operands.append(operand)
            postFix += operand + " "
            return True
        else:
            return False

    def add_operator(operator):
        nonlocal stack, postFix
        if operator in all_operators:
            if operator == '(':
                stack.append(operator)
            elif operator == ')':
                stackOperator = stack.pop()
                while stackOperator != '(':
                    postFix += stackOperator + " "
                    stackOperator = stack.pop()
            else:
                operatorPrecedence = all_operators[operator]["precedence"]
                stackOperator = stack.pop()
                while all_operators[stackOperator]["precedence"] >= operatorPrecedence:
                    postFix += stackOperator + " "
                    stackOperator = stack.pop()
                stack.append(stackOperator)
                stack.append(operator)
            return True
        else:
            return False

    if len(exp) == 0:
        handle_invalid_scenario("Invalid Expression: " + exp)

    if len(exp) > 0 and exp[-1] in all_operators and exp[-1] != ')':
        handle_invalid_scenario(
            "Expression ends with an operator, which is invalid. System will exit"
        )
    exp += ')'
    for letter in exp:
        if letter in all_operators:
            #Some operators can be more than one letter. So, checking if that is an operator
            check = add_operand() or add_operator(operand)
            if not check:
                handle_invalid_scenario("Invalid input " + operand)
            operand = ""
            add_operator(letter)
        else:
            operand += letter
    add_operand()
    postFix= postFix.replace(",", "")
    postFix= postFix.replace("  ", " ")
    return postFix



from math import *

trigonometric = {
    "sin": lambda arg: sin(arg[0]),
    "cos": lambda arg: cos(arg[0]),
    "tan": lambda arg: tan(arg[0]),
    "asin": lambda arg: asin(arg[0]),
    "acos": lambda arg: acos(arg[0]),
    "atan": lambda arg: atan(arg[0]),
    "sinh": lambda arg: sinh(arg[0]),
    "cosh": lambda arg: cosh(arg[0]),
    "tanh": lambda arg: tanh(arg[0]),
    "asinh": lambda arg: asinh(arg[0]),
    "acosh": lambda arg: acosh(arg[0]),
    "atanh": lambda arg: atanh(arg[0]),
    "hypot": lambda arg: hypot(arg[1], arg[0]),
    "deg": lambda arg: degrees(arg[0]),
    "rad": lambda arg: radians(arg[0])
}


from .arithmetic_operations import basic_arithmetic
from .algebraic_operations import algebraic
from .trigonometric import trigonometric
from .logical import logical
from math import *
from datetime import *
from .handle_invalid_scenarios import ParserError

default_variables = {
    "pi": pi,
    "e": e,
    "tau": 2 * pi,
    "True": True,
    "False": False,
    "now": datetime.now(),
    "today": date.today()
}


def perform_operation(operator, arg):
    all_operations = {**basic_arithmetic, **algebraic,  **trigonometric, **logical}
    try:
        operation = all_operations.get(operator, None)
        return all_operations.get(operator, None)(arg)
    except Exception as e:
        raise ParserError(str(e))